#ifndef safe_mutex_hh
#define safe_mutex_hh

#include <mutex>
#include <condition_variable>

// clang has -Wthread-safety but the std mutex header doesn't 
// have the necessary annotations.  So I've taken this verbatim from 
// the clang web page for the feature (http://clang.llvm.org/docs/ThreadSafetyAnalysis.html) 
// and modified it so that the Mutex and MutexLocker classes are renamed 
// safe::mutex and safe::unique_lock, for consistency with the std:: classes.
// I've also removed some deprecated attributes. 

// Code should use safe::mutex/unique_lock/condition_variable.  These just 
// use the std::versions directly on other compilers.  On clang they are thin 
// wrappers that add the required attributes.  There should be no run-time cost.
// Data that is protected by the mutex should be annotated using GUARDED_BY().


// Enable thread safety attributes only with clang.
#if defined(__clang__) && (!defined(SWIG))
#define THREAD_ANNOTATION_ATTRIBUTE__(x)   __attribute__((x))
#define DECLARE_SAFE_MUTEX
#else
#define THREAD_ANNOTATION_ATTRIBUTE__(x)   // no-op
#endif

#define CAPABILITY(x) \
  THREAD_ANNOTATION_ATTRIBUTE__(capability(x))

#define SCOPED_CAPABILITY \
  THREAD_ANNOTATION_ATTRIBUTE__(scoped_lockable)

#define GUARDED_BY(x) \
  THREAD_ANNOTATION_ATTRIBUTE__(guarded_by(x))

#define PT_GUARDED_BY(x) \
  THREAD_ANNOTATION_ATTRIBUTE__(pt_guarded_by(x))

#define ACQUIRED_BEFORE(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(acquired_before(__VA_ARGS__))

#define ACQUIRED_AFTER(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(acquired_after(__VA_ARGS__))

#define REQUIRES(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(requires_capability(__VA_ARGS__))

#define REQUIRES_SHARED(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(requires_shared_capability(__VA_ARGS__))

#define ACQUIRE(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(acquire_capability(__VA_ARGS__))

#define ACQUIRE_SHARED(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(acquire_shared_capability(__VA_ARGS__))

#define RELEASE(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(release_capability(__VA_ARGS__))

#define RELEASE_SHARED(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(release_shared_capability(__VA_ARGS__))

#define TRY_ACQUIRE(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(try_acquire_capability(__VA_ARGS__))

#define TRY_ACQUIRE_SHARED(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(try_acquire_shared_capability(__VA_ARGS__))

#define EXCLUDES(...) \
  THREAD_ANNOTATION_ATTRIBUTE__(locks_excluded(__VA_ARGS__))

#define ASSERT_CAPABILITY(x) \
  THREAD_ANNOTATION_ATTRIBUTE__(assert_capability(x))

#define ASSERT_SHARED_CAPABILITY(x) \
  THREAD_ANNOTATION_ATTRIBUTE__(assert_shared_capability(x))

#define RETURN_CAPABILITY(x) \
  THREAD_ANNOTATION_ATTRIBUTE__(lock_returned(x))

#define NO_THREAD_SAFETY_ANALYSIS \
  THREAD_ANNOTATION_ATTRIBUTE__(no_thread_safety_analysis)


namespace safe {

#ifdef DECLARE_SAFE_MUTEX

template <typename M>
class SCOPED_CAPABILITY unique_lock;


class CAPABILITY("mutex") mutex
{
  typedef std::mutex mut_t;
  mut_t mut;

public:
  void lock() ACQUIRE() { mut.lock(); }
  void unlock() RELEASE() { mut.unlock(); }

  // For negative capabilities.
  const mutex& operator!() const { return *this; }

  friend class unique_lock<mutex>;
};


template <typename M>
class SCOPED_CAPABILITY unique_lock
{
  // This accesses the mutex's mut member directly so that the lock is 
  // really a std::unique_lock<std::mutex>; if it isn't, the condition 
  // variable doesn't work.

  typedef std::unique_lock<typename M::mut_t> std_lock_t;
  std_lock_t std_lock;

public:
  unique_lock(M& m) ACQUIRE(m):
    std_lock(m.mut)
  {}

  ~unique_lock() RELEASE()
  {}

  friend class condition_variable;
};


class condition_variable
{
  std::condition_variable cv;

public:

  void notify_all()
  {
    cv.notify_all();
  }

  void wait(unique_lock<mutex>& l)
  {
    cv.wait(l.std_lock);
  }

  template<typename REL_TIME_T>
  std::cv_status wait_for( unique_lock<mutex>& l,
                           REL_TIME_T rel_time)
  {
    return cv.wait_for(l.std_lock,rel_time);
  }

};

#else

using std::mutex;
using std::unique_lock;
using std::condition_variable;

#endif

};  // namespace safe


#endif  // safe_mutex_hh



